/*-
 * Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: tree.c,v 1.16 2016/02/25 22:20:47 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <babolo/BLINflag.h>
#include "deleted.h"

#define RES_LOW(A)                                                                                    \
    do {                                                                                              \
        ifBLIN_QX1("Resources low in ###%03d", (A));                                                  \
        parm->conf->res_low = true;                                                                   \
        goto brk;                                                                                     \
    } while(0)

#define DUMPT(C,T)                                                                                    \
    do {                                                                                              \
        localtime_r(&(T), &t);                                                                        \
        ifBLIN_QX4( C" %04d-%02d-%02d %02d:%02d:%02d"                                                 \
                  , 1900 + t.tm_year, 1 + t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec          \
                  );                                                                                  \
    } while(0)

static void
debug_tree(d_entry *tree, int level) {
    d_entry *l;
    int i;

    for (l = tree; l; l = l->next) {
        for (i = 0; i < level; i++) fputc('|', stderr);
        if  (l) {
            debug_date(stderr, l->time);
            fprintf( stderr, " %"BLIN_X" next %"BLIN_X" high %"BLIN_X" %s\n"
                   , BLIN_I(l), BLIN_I(l->next), BLIN_I(l->high), l->name
                   );
        } else {
            fprintf( stderr, "-----------------\n");
        }
        if  (l->high) debug_tree(l->high, level + 1);
}   }

static void
free_node(d_entry *node) {
    if  (node && node->name) free(node->name);
    free(node);
}

void
free_tree(d_entry *tree) {
    d_entry *d;

    while (tree) {
        if  (tree->high && tree->next) {
            d = tree->next;
            if  (d->high && d->next) {
                tree->next = d->high;
                d->high = tree;
                tree = d;
            } else if (d->high) {
                tree->next = d->high;
                free_node(d);
            } else if (d->next) {
                tree->next = d->next;
                free_node(d);
            } else {
                tree->next = NULL;
                free_node(d);
            }
        } else if (tree->high) {
            d = tree;
            tree = tree->high;
            free_node(d);
        } else if (tree->next) {
            d = tree;
            tree = tree->next;
            free_node(d);
        } else {
            free_node(tree);
            tree = NULL;
}   }   }

static time_t
youngest(time_t *a, time_t b) {
    if  ((!!b) && (*a < b)) *a = b;
    return(*a);
}

static time_t
select_time(deleted_pid *conf, struct stat *sb, aim *mask) {
    time_t result = 0;
    bool defined;

#   define blin_internal_flags (conf->flags & BLIN_MASK)
    defined = false;
    if  (   (mask->mask & (USE_O | USE_G | USE_R))
        &&  (!(mask->mask & USE_O) || sb->st_uid == mask->own)
        &&  (!(mask->mask & USE_G) || sb->st_gid == mask->grp)
        &&  (!(mask->mask & USE_R) || sb->st_mode == mask->mod)
        ) {
        result = NOW;
        defined = true;
    }
    if  (mask->mask & USE_A) {
        result = sb->st_atime;
        defined = true;
    }
    if  (mask->mask & USE_C) {
        if  (!defined) {
            result = sb->st_ctime;
        } else if (result < sb->st_ctime) {
            result = sb->st_ctime;
        }
        defined = true;
    }
    if  (mask->mask & USE_M) {
        if  (!defined) {
            result = sb->st_mtime;
        } else if (result < sb->st_mtime) {
            result = sb->st_mtime;
        }
        defined = true;
    }
    if  (mask->mask & USE_B) {
        if  (!defined) {
            result = sb->st_birthtime;
        } else if (result < sb->st_birthtime) {
            result = sb->st_birthtime;
        }
        defined = true;
    }
    result += mask->ttl;
    if  (!defined) {
        result = 0;
        ifBLIN_QX0("time mask is empty");
    }
    return(result);
#   undef blin_internal_flags
}

static char *
high_name(const char *path, const char *next) {
    char *next_name;

    asprintf(&next_name, "%s%s%s", path, (*path && path[strlen(path) - 1] == '/') ? "" : "/", next);
    return(next_name);
}

static d_entry *
create_n(const char *name, time_t time) {
    d_entry *d;

    d = malloc(sizeof(d_entry));
    if  (d) {
        d->high = NULL;
        d->next = NULL;
        d->name = name ? strdup(name) : NULL;
        d->time = time;
    }
    return(d);
}

static void
insert_n(d_entry **list, d_entry *d) {
    d_entry **l;
    int i;

    for (i = 0, l = list; *l && ((*l)->time < d->time); l = &((*l)->next)) ++i;
    if  (i) FREE_TREE(d->high);
    if  (*l) FREE_TREE((*l)->high);
    d->next = *l;
    *l = d;
}

static d_entry **
isin_tree(d_entry **list, const char *name) {
    d_entry **d;

    for (d = list; !!d && !!*d; d = &(*d)->next) if (!strcmp(name, (*d)->name)) return(d);
    return(NULL);
}

static void
delete_n(d_entry **list) {
    d_entry *d;

    if  (!!list) {
        d = *list;
        if  (!!d) {
            FREE_TREE(d->high);
            *list = d->next;
            d->next = NULL;
            FREE_TREE(d);
}   }   }

static int
bubble_n(d_entry **list) {
    d_entry *d, **l;
    int i;

    i = 0;
    if  (!!list && !!*list && !!(*list)->next) {
        d = *list;
        for (l = list; !!*(l = &((*l)->next)) && ((*l)->time < d->time);) ++i;
        if  (i) {
            FREE_TREE(d->high);
            *list = d->next;
            d->next = *l;
            *l = d;
    }   }
    return(i);
}

static DIR *
atdir(deleted_pid *parm, const char *path, struct stat *sb, int *df) {
    DIR *dirp = NULL;
    struct stat sb1;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    if  (!!lstat(path, sb)) {
        ifBLIN_QW0("lstat %s~", path);
        goto out;
    }
    if  (!S_ISDIR(sb->st_mode)) {
        ifBLIN_QX0("not a dir %s~", path);
        errno = ENOTDIR;
        goto out;
    }
    if  (!(dirp = opendir(path))) {
        ifBLIN_QW0("opendir %s~", path);
        goto out;
    }
    if  (0 > (*df = dirfd(dirp))) {
        ifBLIN_QW0("dirfd %s~", path);
        goto out;
    }
    if  (fstat(*df, &sb1)) {
        ifBLIN_QW0("fstat %s~", path);
        goto out;
    }
    if  ((sb->st_dev != sb1.st_dev) || (sb->st_ino != sb1.st_ino)) {
        ifBLIN_QX0("subst %s~", path);
        errno = EIDRM;
        goto out;
    }
    return(dirp);
out:
    if  (!!dirp) closedir(dirp);
    return(NULL);
#   undef blin_internal_flags
}

static time_t
do_lnk(deleted_pid *parm, time_t now, struct stat *sb, const char *name) {
    time_t unsleep;
    struct tm t;
    int f;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    unsleep = select_time(parm->conf, sb, &parm->conf->lmask);
    DUMPT("LNK 0   ", unsleep);
    f = 0;
    if  (parm->conf->lmask.mask & (USE_P | USE_L)) {
        /* XXXX     readlink() */
        if  (!!stat(name, sb)) {
            ifBLIN_QX1("stat %s~", name);
            f = -1;
        } else if (parm->conf->lmask.mask & USE_P) {
            if  (S_ISDIR(sb->st_mode)) {
                youngest(&unsleep, select_time(parm->conf, sb, &parm->conf->dmask));
                DUMPT("LNK DIR ", unsleep);
            } else if (S_ISREG(sb->st_mode)) {
                youngest(&unsleep, select_time(parm->conf, sb, &parm->conf->fmask));
                DUMPT("LNK REG ", unsleep);
            } else {
                unsleep = now + parm->conf->lmask.ttl;
                DUMPT("LNK LNK ", unsleep);
    }   }   }
    unsleep += parm->conf->lmask.ttl;
    DUMPT("LNK 1   ", unsleep);
    if  (!f && ((unsleep + BONUS_LIFE) <= now) && (parm->conf->lmask.mask & USE_L)) {
        if  (0 > (f = open(name, O_RDONLY | O_EXLOCK | O_NONBLOCK))) {
            ifBLIN_QW1("open %s~", name);
            if  ((errno == EMFILE) || (errno == ENFILE)) {
                RES_LOW(500);                                                                   /*###*/
            } else {
                unsleep = now + parm->conf->lmask.ttl;
                DUMPT("LNK lock", unsleep);
            }
        } else {
            close(f);
    }   }
    if  ((unsleep + BONUS_LIFE) <= now) {
        if  (!!unlink(name)) {
            ifBLIN_QW1("unlink LNK %s", name);
            unsleep = now + parm->conf->lmask.ttl;
            DUMPT("LNK rm- ", unsleep);
        } else {
            unsleep = 0;
            ifBLIN_QX4("LNK rm+");
    }   }
brk:
    return(unsleep);
#   undef blin_internal_flags
}

static time_t
do_reg(deleted_pid *parm, time_t now, struct stat *sb, const char *name) {
    struct stat sb1;
    time_t unsleep;
    struct tm t;
    int f;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    unsleep = select_time(parm->conf, sb, &parm->conf->fmask) + parm->conf->fmask.ttl;
    DUMPT("REG 0   ", unsleep);
    if  (((unsleep + BONUS_LIFE) <= now) && (parm->conf->fmask.mask & USE_L)) {
        if  (0 > (f = open(name, O_RDONLY | O_EXLOCK | O_NONBLOCK))) {
            ifBLIN_QW1("open %s~", name);
            if  ((errno == EMFILE) || (errno == ENFILE)) {
                RES_LOW(502);                                                                   /*###*/
            } else {
                unsleep = now + parm->conf->fmask.ttl;
                DUMPT("REG lock", unsleep);
            }
        } else {
            if  (!!fstat(f, &sb1)) {
                ifBLIN_QW1("fstat %s~", name);
                unsleep = now + parm->conf->fmask.ttl;
                DUMPT("REG stat", unsleep);
            } else if ((sb->st_dev != sb1.st_dev) || (sb->st_ino != sb1.st_ino)) {
                ifBLIN_QX1( "d%u<>%u i%u<>%u %s~"
                          , sb->st_dev, sb1.st_dev, sb->st_ino, sb1.st_ino, name
                          );
                unsleep = now + parm->conf->fmask.ttl;
                DUMPT("REG diff", unsleep);
            }
            close(f);
    }   }
    if  ((unsleep + BONUS_LIFE) <= now) {
        if  (!!unlink(name)) {
            ifBLIN_QW1("unlink REG %s", name);
            unsleep = now + parm->conf->fmask.ttl;
            DUMPT("REG rm- ", unsleep);
        } else {
            unsleep = 0;
            ifBLIN_QX4("REG rm+");
    }   }
brk:
    return(unsleep);
#   undef blin_internal_flags
}

static time_t
do_dir(deleted_pid *parm, time_t now, struct stat *sb, const char *name) {
    struct stat sb1;
    time_t unsleep;
    struct tm t;
    int f;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    unsleep = select_time(parm->conf, sb, &parm->conf->dmask) + parm->conf->dmask.ttl;
    DUMPT("DIR 0   ", unsleep);
    if  (!!parm->conf->dmask.ttl) {
        if  (((unsleep + BONUS_LIFE) <= now) && (parm->conf->dmask.mask & USE_L)) {
            if  (0 > (f = open(name, O_RDONLY | O_EXLOCK | O_NONBLOCK))) {
                ifBLIN_QW1("open %s~", name);
                if  ((errno == EMFILE) || (errno == ENFILE)) {
                    RES_LOW(504);                                                               /*###*/
                } else {
                    unsleep = now + parm->conf->dmask.ttl;
                    DUMPT("DIR lock", unsleep);
                }
            } else {
                if  (!!fstat(f, &sb1)) {
                ifBLIN_QW1("fstat %s~", name);
                    unsleep = now + parm->conf->dmask.ttl;
                    DUMPT("DIR stat", unsleep);
                } else if ((sb->st_dev != sb1.st_dev) || (sb->st_ino != sb1.st_ino)) {
                    ifBLIN_QX1( "d%u<>%u i%u<>%u %s~"
                              , sb->st_dev, sb1.st_dev, sb->st_ino, sb1.st_ino, name
                              );
                    unsleep = now + parm->conf->dmask.ttl;
                    DUMPT("DIR diff", unsleep);
                }
                close(f);
        }   }
        if  ((unsleep + BONUS_LIFE) <= now) {
            if  (!!rmdir(name)) {
                ifBLIN_QW1("rmdir %s~", name);
                unsleep = now + parm->conf->dmask.ttl;
                DUMPT("DIR rm- ", unsleep);
            } else {
                unsleep = -1;
                ifBLIN_QX4("DIR rm+");
    }   }   }
brk:
    return(unsleep);
#   undef blin_internal_flags
}

d_entry *
/***********************************************************************
 *                                                                     */
clear_tree(deleted_pid *parm, const char *path, d_entry *tree) {
/*                                                                     *
 ***********************************************************************
 * Build partial tree
 *              high       high       high       high
 * tree -> node ----> node ----> node ----> node ----> node  < the oldest node in the tree
 *          |          |          |                     |
 *          | next     | next     | next                | next
 *          V          V          V                     V
 *         node       node       node                  node
 *          |          |          |                     |
 *          | next     | next     | next                | next
 *          V          V          V                     V
 *         node       node       node                  node
 *          |          |                                |
 *          | next     | next   sorted                  | next
 *          V          V        by time                 V
 *         node       node                             node
 *
 *        sorted     sorted                           sorted
 *        by time    by time                          by time - oldest is highest,
 *                                                              younger is lower,
 *                                                              bubbles go down (in sort)
 *
 * Next link points to the node in the same directory as parent.
 * Next linked list sorted by time.
 * High link points to subdirectory.
 * The only high linked line exists - the way to the
 * oldest node in the directory.
 * Time of the directory is oldest time of files and subdirectories
 * of this directory if it is not empty.
 **********************************************************************
 */
    time_t now, unsleep;
    d_entry *d, **node;
    struct dirent *dp;
    char *next_name;
    struct stat sb;
    DIR *dirp;
    size_t s;
    int df;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    ifBLIN_QX2("+ %"BLIN_X" %s~", tree, path);
    if  (!(dirp = atdir(parm, path, &sb, &df))) {
        ifBLIN_QW0("atdir");
        goto out;
    }
    while (!!tree) {
    ;   now = NOW;
    ;   if  (!!fchdir(df)) {
    ;       ifBLIN_QW0("fchdir");
    ;       goto out;
    ;   }
    ;   if  (!!lstat(tree->name, &sb)) {
    ;       ifBLIN_QX1(", lstat %s~", tree->name);
    ;       delete_n(&tree);
    ;       continue;
    ;   }
    ;   if  ((tree->time + BONUS_LIFE) > now) goto brc;
    ;   switch (sb.st_mode & S_IFMT) {
        case S_IFLNK:
    ;       ifBLIN_QX5(", symlink=%s~", tree->name);
    ;       if  (!!parm->conf->lmask.ttl) {
    ;       ;   unsleep = do_lnk(parm, now, &sb, tree->name);
    ;       ;   if  (!unsleep) {
    ;       ;       delete_n(&tree);
    ;       ;   } else {
    ;       ;       tree->time = unsleep;
    ;       ;       if  (!bubble_n(&tree)) goto brc;
    ;       ;   }
    ;       ;   ifBLIN_QX5(", symlink %s~", tree->name);
    ;       } else {
    ;           delete_n(&tree);
    ;       }
    ;       ifBLIN_QX5(", symlink %s~", tree->name);
    ;       continue;
    ;   case S_IFREG:
    ;       ifBLIN_QX5(", file=%s~", tree->name);
    ;       if  (!!parm->conf->fmask.ttl) {
    ;       ;   unsleep = do_reg(parm, now, &sb, tree->name);
    ;       ;   if  (!unsleep) {
    ;       ;       delete_n(&tree);
    ;       ;   } else {
    ;       ;       tree->time = unsleep;
    ;       ;       if  (!bubble_n(&tree)) goto brc;
    ;       ;   }
    ;       ;   ifBLIN_QX5(", file %s~", tree->name);
    ;       } else {
    ;           delete_n(&tree);
    ;       }
    ;       ifBLIN_QX5(", file %s~", tree->name);
    ;       continue;
    ;   case S_IFDIR:
    ;       ifBLIN_QX5(", dir=%s~", tree->name);
    ;       unsleep = do_dir(parm, now, &sb, tree->name);
    ;       if  (0 > unsleep) {
    ;           delete_n(&tree);
    ;       } else {
    ;           if  (!(next_name = high_name(path, tree->name))) RES_LOW(704);                  /*###*/
    ;           tree->high = clear_tree(parm, next_name, tree->high);
    ;           free(next_name);
    ;           if  (!!tree->high && !!tree->high->time) {
    ;               tree->time = tree->high->time;
    ;           } else {
    ;               tree->time = now + parm->conf->wait_up;
    ;           }
    ;           if  (!bubble_n(&tree)) goto brc;
    ;           ifBLIN_QX5(", dir %s~", tree->name);
    }   }   }
brc:
    s = strlen(parm->pidname);
    while ((dp = readdir(dirp)) != NULL) {
    ;   now = NOW;
    ;   if  (   !strcmp(dp->d_name, ".")
            ||  !strcmp(dp->d_name, "..")
            ||  (   dp->d_namlen = s
                &&  !strcmp(path, "/")
                &&  !strncmp(dp->d_name, parm->pidname, s)
            )   ) {
    ;       ifBLIN_QX5("; skip %s~", dp->d_name);
    ;       continue;
    ;   }
    ;   if  (!!fchdir(df)) {
    ;       ifBLIN_QW0("fchdir");
    ;       FREE_TREE(tree);
    ;       break;
    ;   }
    ;   if  (!!lstat(dp->d_name, &sb)) {
    ;       ifBLIN_QX1(", lstat %s~", dp->d_name);
    ;       continue;
    ;   }
    ;   switch (sb.st_mode & S_IFMT) {
        case S_IFLNK:
    ;       ifBLIN_QX5("; symlink=%s~", dp->d_name);
    ;       if  (!!parm->conf->lmask.ttl) {
    ;       ;   node = isin_tree(&tree, dp->d_name);
    ;       ;   if  (!node) {
    ;       ;       unsleep = do_lnk(parm, now, &sb, dp->d_name);
    ;       ;       if  (!unsleep) {
    ;       ;           delete_n(node);
    ;       ;       } else {
    ;       ;           if  (!(d = create_n(dp->d_name, unsleep))) RES_LOW(601);                /*###*/
    ;       ;           insert_n(&tree, d);
    ;       ;       }
    ;       ;       ifBLIN_QX5("; symlink %s~", dp->d_name);
    ;       }   }
    ;       continue;
    ;   case S_IFREG:
    ;       ifBLIN_QX5("; file=%s~", dp->d_name);
    ;       if  (!!parm->conf->fmask.ttl) {
    ;       ;   node = isin_tree(&tree, dp->d_name);
    ;       ;   if  (!node) {
    ;       ;       unsleep = do_reg(parm, now, &sb, dp->d_name);
    ;       ;       if  (!unsleep) {
    ;       ;           delete_n(node);
    ;       ;       } else {
    ;       ;           if  (!(d = create_n(dp->d_name, unsleep))) RES_LOW(603);                /*###*/
    ;       ;           insert_n(&tree, d);
    ;       ;       }
    ;       ;       ifBLIN_QX5("; file %s~", dp->d_name);
    ;       }   }
    ;       continue;
    ;   case S_IFDIR:
    ;       ifBLIN_QX5("; dir=%s~", dp->d_name);
    ;       if  (!(node = isin_tree(&tree, dp->d_name))) {
    ;       ;   unsleep = do_dir(parm, now, &sb, dp->d_name);
    ;       ;   if  (unsleep < (now + parm->conf->wait_up)) unsleep = now + parm->conf->wait_up;
    ;       ;   ifBLIN_QX6("   unsleep %d=%s~", unsleep, dp->d_name);
    ;       ;   if  (0 <= unsleep) {
    ;       ;       ifBLIN_QX6("   create %s~", dp->d_name);
    ;       ;       d = create_n(dp->d_name, unsleep);
    ;       ;       if  (!d) RES_LOW(605);                                                      /*###*/
    ;       ;       if  (!(next_name = high_name(path, dp->d_name))) RES_LOW(606);              /*###*/
    ;       ;       d->high = clear_tree(parm, next_name, NULL);
    ;       ;       free(next_name);
    ;       ;       if  (!!d->high && !!d->high->time) d->time = d->high->time;
    ;       ;       insert_n(&tree, d);
    ;       ;   }
    ;       ;   ifBLIN_QX5("; dir %s~", dp->d_name);
    ;       }
    ;       continue;
    }   }
brk:
    closedir(dirp);
    if  (!!chdir(path)) {
        ifBLIN_QW0("chdir %s~", path);
    }
out:
    ifBLIN_QO3 debug_tree(tree, 0);
    ifBLIN_QX2("- %"BLIN_X" %s~", tree, path);
    if  (!tree) FREE_TREE(tree);
    return(tree);
#   undef blin_internal_flags
}
